#include <TinyWireS.h>
#include <EEPROM.h>

union Buffer
{
   long longNumber;
   byte longBytes[4];
};
Buffer bPH; // pH
Buffer bRead1; // point 1 pH reading during calibration in mV
Buffer bRead2; // point 2 pH reading during calibration in mV
Buffer bRead3; // point 3 pH reading during calibration in mV
Buffer bBuffer1; // buffer 1 solution in pH
Buffer bBuffer2; // buffer 2 solution in pH
Buffer bBuffer3; // buffer 3 solution in pH
Buffer bSlope1; // low end slope percentage
Buffer bSlope2; // high end slope percentage
Buffer bOffset1; // point 1 calibration offset in mV
Buffer bTempC; // sample temperature
Buffer bCalTempC; // calibration solution temperature

byte SendCount = 0;  // used to count bytes being sent to master device
uint8_t sendByte;  // passes byte to master device

#define PROBE_IN A3 //57   // anlaog a3  - green wire
#define PROBE_REF A2 //56   // anlaog a2  - white wire

//#define PROBE_MV_TO_PH 59.16
//#define TEMP_CORRECTION_FACTOR 0.03

byte ver;  // eeprom 0
//byte i2cAdd = 30;  // eeprom 1
byte i2cAdd = 40;  // eeprom 1
float cal_mV1 = -0;       // point 1 pH reading during calibration - eeprom 2
float cal_mV2 = 177.48;  // point 2 pH reading during calibration - eeprom 6
float cal_mV3 = -177.48;  // point 3 pH reading during calibration - eeprom 10

float buffer1; // buffer 1 solution - eeprom 14
float buffer2; // buffer 2 solution - eeprom 18
float buffer3; // buffer 3 solution - eeprom 22

byte calType;  // type of calibration being used, 1point, 2point or 3point - eeprom 26

byte calibrating = 0;  // disables calibration while calibrating
byte calibrationEnabled;  // tracks if calibration is enabled disabled
int incoming = 0;
float pH;
float slope = 0;
float referenceRange = 0;
float readingRange = 0;

// variables below are calculated in sketch
double mVc;  // mV calibrated
float pHc;  // pH calibrated
float mc1; //slope correction 4-7
float bc1; //offset correction 4-7
float m1; //slope for 4-7
float b1; //offset for 4-7
float mc2; //slope correction 7-10
float bc2; //offset correction 7-10
float m2; //slope for 7-10
float b2; //offset for 7-10
float mVPerPH;  // calibrated mV per pH without temperature compensation
float temp_mvPerPH;  // calibrated mV per pH with temperature compensation

float tempCalibrated;  // temperature of calibration solution during calibration
float tempC = 25.0;           // temperature of sample being monitored in Celsius

byte g = 0;
void setup()
{
  byte z = 255;
  EEPROM.get(50, z);
//    EEPROM.put(1, 40);  // I2C address - 40, 41, 42, 43, 44, 45, 46, 47
  
  if (z == 255) 
  {
  //  Defaults
  
    EEPROM.put(0, 1);  // version
//    EEPROM.put(1, 30);  // I2C address - 30, 31, 32, 33, 34, 35, 36, 37
    EEPROM.put(1, 40);  // I2C address - 40, 41, 42, 43, 44, 45, 46, 47
    EEPROM.put(2, 0.0);  // cal_mV1
    EEPROM.put(6, 177.48);  // cal_mV2
    EEPROM.put(10, -177.48);  // 
    EEPROM.put(14, 0.0);  // buffer1
    EEPROM.put(18, 177.48);  // buffer2
    EEPROM.put(22, -177.48);  // buffer3
    EEPROM.put(26, 2);  // calibration type
    EEPROM.put(27, 1);  // calibration enabled/disabled
    EEPROM.put(28, 25.0);  // calibration temperature
    EEPROM.put(50, 1);  // version
  }
  pinMode(1, OUTPUT);
  digitalWrite(1, HIGH);
  
  EEPROM.get(0, ver);
  EEPROM.get(1, i2cAdd);
  EEPROM.get(2, cal_mV1);  // low
  EEPROM.get(6, cal_mV2);  // mid
  EEPROM.get(10, cal_mV3);  // high
  EEPROM.get(14, buffer1);
  EEPROM.get(18, buffer2);
  EEPROM.get(22, buffer3);
  EEPROM.get(26, calType);
  EEPROM.get(27, calibrationEnabled);
  EEPROM.get(28, tempCalibrated);
  calculateSlopes();
//  measuremV();
  TinyWireS.begin(i2cAdd);
  TinyWireS.onReceive(receiveEvent); // register event
  TinyWireS.onRequest(requestEvent); // register event
}
void loop()
{
  TinyWireS_stop_check();
}
void measuremV() 
{
  digitalWrite(1, LOW);
  double differentialmV;
  float  probeIn, probeRef, inputV, adcResolution;
  analogReference(DEFAULT);
  inputV = getVin();
  probeIn  = ((readADC(PROBE_IN) * inputV) / 1024.0);
  probeRef = ((readADC(PROBE_REF) * inputV) / 1024.0);
  differentialmV = (probeIn - probeRef);

  if (differentialmV <= -1.1)
  {
//    analogReference(INTERNAL2V56);
    analogReference(INTERNAL);
    adcResolution = 2.56;
  }
  else if (differentialmV <= -0)
  {
    analogReference(INTERNAL);  // set to INTERNAL for ATtiny
    adcResolution = 1.1;
  }
  else if (differentialmV <= 1.1)
  {
    analogReference(INTERNAL);
    adcResolution = 1.1;
  }
  else if (differentialmV <= 2.56)
  {
    analogReference(INTERNAL);
    adcResolution = 2.56;
  }

  // take a reading
  probeIn  = ((readADC(PROBE_IN) * adcResolution) / 1024.0);
  probeRef = ((readADC(PROBE_REF) * adcResolution) / 1024.0);

  // turn V into mV  
  differentialmV = (probeIn - probeRef) * 1000;
  temp_mvPerPH = 59.16;
  mVPerPH = 59.16;
  if ((calibrating == 0) && (calibrationEnabled == 1))
  {
    switch (calType)
    {
      case 1:  // 1 point calibration being used
        pHc = mc1*differentialmV + bc1;
        differentialmV = (7-pHc)*mVPerPH;
      break;

      case 2:  // 2 point calibration being used
        pHc = mc1*differentialmV + bc1;
        mVPerPH = (cal_mV2 - (cal_mV1)) / 3;
        differentialmV = (7.0-pHc)*mVPerPH;
      break;
        
      case 3:  // 3 point calibration being used
        if (differentialmV <= 0.0)  // above 7pH
        {
          pHc = mc2*differentialmV + bc2;
          mVPerPH = (cal_mV1 - (cal_mV3)) / 3;
          differentialmV = (7-pHc)*mVPerPH;

        }
        else  // below 7pH
        {
          pHc = mc1*differentialmV + bc1;
          mVPerPH = (cal_mV2 - (cal_mV1)) / 3;
          differentialmV = (7-pHc)*mVPerPH;
        }
      break;
    }
//  following link used to calculate temperature compsenation based on calibrated temp and sample temp
//  https://www.hach.com/asset-get.download.jsa?id=17525673904

//  follwing link has a pH/mV ratio chart which was used for comparisions to verify temperature compensation
//  https://web-material3.yokogawa.com/TNA0924.us.pdf

    // add temperature compensation
    temp_mvPerPH = mVPerPH * (tempC+273.15) / ((tempCalibrated)+273.15);
  }

  // convert to pH
  pH = fabs(7.0 - (differentialmV / temp_mvPerPH));
  bPH.longNumber = pH * 10000;
  SendCount = 0;
  digitalWrite(1, HIGH);
}
void requestEvent()  // send data to controller
{
//  if (SendCount == 4) SendCount = 0;

  switch (incoming)
  {
    case 1: 
      if (pH == 0.000) break;  // don't send if pH is 0.000
      switch (SendCount)
      {
        case 0:
          sendByte = bPH.longBytes[0];
          break;
        case 1:
          sendByte = bPH.longBytes[1];
          break;
        case 2:
          sendByte = bPH.longBytes[2];
          break;
        case 3:
          sendByte = bPH.longBytes[3];
          break;
      }
      TinyWireS.send(sendByte);
      SendCount++;
      if (SendCount > 3) SendCount = 0;
    break;
        
    case 3: TinyWireS.send(ver); break;  // send circuit version
    case 9: TinyWireS.send(i2cAdd); break;  // get I2C address
    
    case 11:  // send low slope
      switch (SendCount)
      {
        case 0:
          sendByte = bSlope1.longBytes[0];
          break;
        case 1:
          sendByte = bSlope1.longBytes[1];
          break;
        case 2:
          sendByte = bSlope1.longBytes[2];
          break;
        case 3:
          sendByte = bSlope1.longBytes[3];
          break;
      }
      TinyWireS.send(sendByte);
      SendCount++;
      if (SendCount > 3) SendCount = 0;
    break;
    
    case 12:  // send high slope
      switch (SendCount)
      {
        case 0:
          sendByte = bSlope2.longBytes[0];
          break;
        case 1:
          sendByte = bSlope2.longBytes[1];
          break;
        case 2:
          sendByte = bSlope2.longBytes[2];
          break;
        case 3:
          sendByte = bSlope2.longBytes[3];
          break;
      }
      TinyWireS.send(sendByte);
      SendCount++;
      if (SendCount > 3) SendCount = 0;
    break;
    
    case 13:  // send calibration point 1 offset
      switch (SendCount)
      {
        case 0:
          sendByte = bOffset1.longBytes[0];
          break;
        case 1:
          sendByte = bOffset1.longBytes[1];
          break;
        case 2:
          sendByte = bOffset1.longBytes[2];
          break;
        case 3:
          sendByte = bOffset1.longBytes[3];
          break;
      }
      TinyWireS.send(sendByte);
      SendCount++;
      if (SendCount > 3) SendCount = 0;
    break;
  }
}
void receiveEvent(uint8_t howMany)  // controller is asking for something
{
  SendCount = 0;
  incoming = TinyWireS.receive();
  switch (incoming)
  {
    case 1: SendCount = 0; measuremV(); break;  // get pH
    
    case 2:  // update sample temperature
      bTempC.longBytes[0] = TinyWireS.receive();
      bTempC.longBytes[1] = TinyWireS.receive();
      bTempC.longBytes[2] = TinyWireS.receive();
      bTempC.longBytes[3] = TinyWireS.receive();
      tempC = bTempC.longNumber * 0.0001;
    break;
      
    case 4:  // start calibration
      calibrating = 1;
      calType = TinyWireS.receive();
      
      bCalTempC.longBytes[0] = TinyWireS.receive();
      bCalTempC.longBytes[1] = TinyWireS.receive();
      bCalTempC.longBytes[2] = TinyWireS.receive();
      bCalTempC.longBytes[3] = TinyWireS.receive();
      tempCalibrated = bCalTempC.longNumber * 0.0001;
      EEPROM.put(26, calType);  // calibration type
      EEPROM.put(28, tempCalibrated);  // calibration temp of solution
    break;
    
    case 5:  // reset calibration data
      calibrating = 0;
      calType = 1;
      buffer1 = 0.0;
      cal_mV1 = 0.0;
      buffer2 = 177.48;
      cal_mV2 = 177.48;
      buffer3 = -177.48;
      cal_mV3 = -177.48;
      tempCalibrated = 25.0;
      EEPROM.put(26, calType);  // calibration type
      EEPROM.put(2, cal_mV1);  // cal_mV1
      EEPROM.put(14, buffer1);  // buffer1
      EEPROM.put(6, cal_mV2);  // cal_mV2
      EEPROM.put(18, buffer2);  // buffer2
      EEPROM.put(10, cal_mV3);  // cal_mV3
      EEPROM.put(22, buffer3);  // buffer3
      EEPROM.put(28, tempCalibrated);  // calibration temp of solution
      calculateSlopes();
    break;
    
    case 6:  // change point 1 calibration data
      bBuffer1.longBytes[0] = TinyWireS.receive();
      bBuffer1.longBytes[1] = TinyWireS.receive();
      bBuffer1.longBytes[2] = TinyWireS.receive();
      bBuffer1.longBytes[3] = TinyWireS.receive();
      
      bRead1.longBytes[0] = TinyWireS.receive();
      bRead1.longBytes[1] = TinyWireS.receive();
      bRead1.longBytes[2] = TinyWireS.receive();
      bRead1.longBytes[3] = TinyWireS.receive();
      
      buffer1 = (7 - (bBuffer1.longNumber * 0.0001)) * 59.16;
      cal_mV1 = (7 - (bRead1.longNumber * 0.0001)) * 59.16;
      
      EEPROM.put(2, cal_mV1);  // cal_mV1
      EEPROM.put(14, buffer1);  // buffer1
      if (calType == 1) {calibrating = 0; calculateSlopes();}
    break;
    
    case 7:  // change point 2 calibration data
      bBuffer2.longBytes[0] = TinyWireS.receive();
      bBuffer2.longBytes[1] = TinyWireS.receive();
      bBuffer2.longBytes[2] = TinyWireS.receive();
      bBuffer2.longBytes[3] = TinyWireS.receive();
      
      bRead2.longBytes[0] = TinyWireS.receive();
      bRead2.longBytes[1] = TinyWireS.receive();
      bRead2.longBytes[2] = TinyWireS.receive();
      bRead2.longBytes[3] = TinyWireS.receive();
      
      buffer2 = (7 - (bBuffer2.longNumber * 0.0001)) * 59.16;
      cal_mV2 = (7 - (bRead2.longNumber * 0.0001)) * 59.16;
      
      EEPROM.put(6, cal_mV2);  // cal_mV2
      EEPROM.put(18, buffer2);  // buffer2
      if (calType == 2) {calibrating = 0; calculateSlopes();}
    break;
    
    case 8:  // change point 3 calibration data
      bBuffer3.longBytes[0] = TinyWireS.receive();
      bBuffer3.longBytes[1] = TinyWireS.receive();
      bBuffer3.longBytes[2] = TinyWireS.receive();
      bBuffer3.longBytes[3] = TinyWireS.receive();
      
      bRead3.longBytes[0] = TinyWireS.receive();
      bRead3.longBytes[1] = TinyWireS.receive();
      bRead3.longBytes[2] = TinyWireS.receive();
      bRead3.longBytes[3] = TinyWireS.receive();
      
      buffer3 = (7 - (bBuffer3.longNumber * 0.0001)) * 59.16;
      cal_mV3 = (7 - (bRead3.longNumber * 0.0001)) * 59.16;
      
      EEPROM.put(10, cal_mV3);  // cal_mV3
      EEPROM.put(22, buffer3);  // buffer3
      calibrating = 0;
      calculateSlopes();
    break;
    
    case 10:  // change I2C address
      i2cAdd = TinyWireS.receive();
      TinyWireS.begin(i2cAdd);
      EEPROM.put(1, i2cAdd);  // I2C address
    break;
    
    case 14:  // enable/disable calibration
      calibrationEnabled = TinyWireS.receive();
      EEPROM.put(27, calibrationEnabled);  // calibration enabled/disabled
    break;
    
    case 15:  // change calibration type (not really need but good for service)
      calType = TinyWireS.receive();
      EEPROM.put(26, calType);  // calibration type
    break;
    
    case 16:  // change temperature of calibration solution
      bCalTempC.longBytes[0] = TinyWireS.receive();
      bCalTempC.longBytes[1] = TinyWireS.receive();
      bCalTempC.longBytes[2] = TinyWireS.receive();
      bCalTempC.longBytes[3] = TinyWireS.receive();
      tempCalibrated = bCalTempC.longNumber * 0.0001;
      EEPROM.put(28, tempCalibrated);  // calibration temp of solution
    break;
    
    case 17:  // default settings
      EEPROM.put(0, 1);  // version
//      EEPROM.put(1, 30);  // I2C address - 30, 31, 32, 33, 34, 35, 36, 37
      EEPROM.put(1, 40);  // I2C address - 40, 41, 42, 43, 44, 45, 46, 47
      EEPROM.put(2, 0.0);  // cal_mV1
      EEPROM.put(6, 177.48);  // cal_mV2
      EEPROM.put(10, -177.48);  // 
      EEPROM.put(14, 0.0);  // buffer1
      EEPROM.put(18, 177.48);  // buffer2
      EEPROM.put(22, -177.48);  // buffer3
      EEPROM.put(26, 2);  // calibration type
      EEPROM.put(27, 1);  // calibration enabled/disabled
      EEPROM.put(28, 25.0);  // calibration temperature
      tws_delay(500);
      EEPROM.get(0, ver);
      EEPROM.get(1, i2cAdd);
      EEPROM.get(2, cal_mV1);  // low
      EEPROM.get(6, cal_mV2);  // mid
      EEPROM.get(10, cal_mV3);  // high
      EEPROM.get(14, buffer1);
      EEPROM.get(18, buffer2);
      EEPROM.get(22, buffer3);
      EEPROM.get(26, calType);
      EEPROM.get(27, calibrationEnabled);
      EEPROM.get(28, tempCalibrated);
    break;
  }
}
double readADC(int channel)
{
  uint32_t total = 0UL;
  uint16_t sampleCount = 4096;
  for (uint16_t i = 0; i < sampleCount; i++) {total += analogRead(channel);}
  total = total >> 6;
  double proportional = (total * 1.0) / (0b00000001 << 6);
  return proportional;
}
float getVin()
{
  #if defined(__AVR_ATmega32U4__) || defined(__AVR_ATmega1280__) || \
  defined(__AVR_ATmega2560__)
  ADMUX = _BV(REFS0) | _BV(MUX4) | _BV(MUX3) | _BV(MUX2) | _BV(MUX1);
  #elif defined(__AVR_ATtiny24__) || defined(__AVR_ATtiny44__) || \
  defined(__AVR_ATtiny84__)
  ADMUX = _BV(MUX5) | _BV(MUX0);
  #elif defined(__AVR_ATtiny25__) || defined(__AVR_ATtiny45__) || \
  defined(__AVR_ATtiny85__)
  ADMUX = _BV(MUX3) | _BV(MUX2);
  #else // if defined(__AVR_ATmega32U4__) || defined(__AVR_ATmega1280__) ||
  // defined(__AVR_ATmega2560__)
  ADMUX = _BV(REFS0) | _BV(MUX3) | _BV(MUX2) | _BV(MUX1);
  #endif // if defined(__AVR_ATmega32U4__) || defined(__AVR_ATmega1280__) ||
  // defined(__AVR_ATmega2560__)
  tws_delay(2);
  ADCSRA |= _BV(ADSC);

  while (bit_is_set(ADCSRA, ADSC));

  uint8_t low  = ADCL;
  uint8_t high = ADCH;
  long result = (high << 8) | low;

  result = 1125300L / result;
  return (double(result) - 0) * (6 - 0) / (6000 - 0) + 0;
}
void calculateSlopes()
{
//Slope intercept calculations  
  float BmV1 = 7.0 - (buffer1 / 59.16); //pH for 7pH
  float BmV2 = 7.0 - (buffer2 / 59.16); //pH for 4pH
  float BmV3 = 7.0 - (buffer3 / 59.16); //pH for 10pH

  m1 = (BmV2 - BmV1) / (buffer2 - buffer1);
  b1 = BmV2 - buffer2*m1;
  
  mc1 = (BmV2 - BmV1) / (cal_mV2 - cal_mV1);
  bc1 = BmV2 - cal_mV2*mc1; 
  
  m2 = (BmV1 - BmV3) / (buffer1 - buffer3);
  b2 = BmV3 - buffer3*m2;

  mc2 = (BmV1 - BmV3) / (cal_mV1 - cal_mV3);
  bc2 = BmV3 - cal_mV3*mc2;

  float slopeLow = fabs(m1/mc1)*100;  // low end calibration slope
  float slopeHigh = fabs(m2/mc2)*100;  // high end calibration slope
  
  bSlope1.longNumber = slopeLow * 10000;
  bSlope2.longNumber = slopeHigh * 10000;
//===============================

  // calculate offset percentage from point 1
  float offsetPoint1 = cal_mV1 - buffer1;
  if (offsetPoint1 < 0.0) offsetPoint1 = offsetPoint1 + 1999.0;
  bOffset1.longNumber = offsetPoint1 * 10000;
}
